@@ -4,7 +4,7 @@
 
   Author: Pekka Riikonen <priikone@silcnet.org>
 
-  Copyright (C) 2000 - 2007 Pekka Riikonen
+  Copyright (C) 2000 - 2008 Pekka Riikonen
 
   The contents of this file are subject to one of the Licenses specified 
   in the COPYING file;  You may not use this file except in compliance 
@@ -55,7 +55,7 @@ static void silc_idcache_destructor(SilcIDCache cache,
 				    void *app_context)
 {
   if (cache->destructor)
-    cache->destructor(cache, entry, cache->destructor, app_context);
+    cache->destructor(cache, entry, cache->context, app_context);
 
   memset(entry, 'F', sizeof(*entry));
   silc_free(entry);
